"""
Schema generation utilities for lazy-ninja.
"""
from typing import Type, List, Optional
from pydantic import ConfigDict, create_model, model_validator

from django.db import models
from ninja import Schema

from .base import get_pydantic_type, serialize_model_instance


def generate_schema(
    model: Type[models.Model], 
    exclude: List[str] = None, 
    optional_fields: List[str] = None, 
    update: bool = False
) -> Type[Schema]:
    """
    Generate a Pydantic schema based on a Django model.
    
    Args:
        model: The Django model class
        exclude: A list of field names to exclude from the schema
        optional_fields: A list of field names that should be marked as optional
        update: Whether this is an update schema (all fields optional)
        
    Returns:
        A dynamically created Pydantic schema class
        
    Notes:
        - Fields listed in `optional_fields` or with null=True in the Django model are set as Optional
        - A root validator is added to preprocess the input using `serialize_model_instance`
    """
    exclude = exclude or []
    optional_fields = optional_fields or []
    
    fields = {}
    for field in model._meta.fields:
        if field.name in exclude:
            continue
            
        pydantic_type = get_pydantic_type(field)
        
        if update:
            # For update schemas, all fields are optional
            fields[field.name] = (Optional[pydantic_type], None)
        elif field.name in optional_fields or field.null:
            # Mark field as optional if explicitly specified or Django field allows null
            fields[field.name] = (Optional[pydantic_type], None)
        else:
            # Required field
            fields[field.name] = (pydantic_type, ...)
    
    class DynamicSchema(Schema):
        @model_validator(mode="before")
        def pre_serialize(cls, values):
            """
            Pre-root validator that converts a Django model instance into a dict
            using our serialize_model_instance function.
            """
            if hasattr(values, "_meta"):
                return serialize_model_instance(values)
            return values
        
        model_config = ConfigDict(form_attributes=True)

    schema = create_model(
        model.__name__ + "Schema",
        __base__=DynamicSchema,
        **fields
    )
    
    return schema



import subprocess
import sys
import os


def run_cli_command(args: list[str]) -> subprocess.CompletedProcess:
    env = os.environ.copy()
    env["PYTHONPATH"] = "src"
    return subprocess.run(
        [sys.executable, "-m", "lazy_ninja.cli.client_generator"] + args,
        capture_output=True,
        text=True, 
        env=env
    )
    

def test_help_command():
    result = run_cli_command(["--help"])
    assert result.returncode == 0
    assert "Lazy Ninja CLI" in result.stdout
    assert "generate-client" in result.stdout

    
def test_missing_required_arguments():
    result = run_cli_command(["generate-client", "typescript"])
    assert result.returncode != 0
    assert "--settings" in result.stderr
    
    
def test_invalid_language():
    result = run_cli_command(["generate-client", "invalid", "--settings", "myproject.settings"])
    assert result.returncode == 2
    assert "invalid choice: 'invalid'" in result.stderr
    assert "choose from" in result.stderr


def test_invalid_command():
    result = run_cli_command(["invalid-command"])
    assert result.returncode != 0
    assert "invalid choice" in result.stderr.lower()



import os
import sys
import subprocess
import argparse
import importlib
from pathlib import Path

GENERATOR_CONFIG = {
    "typescript-types": {
        "cmd": ["npx", "openapi-typescript", "{schema}", "--output", "{out}"],
        "ext": "ts",
    },
    "dart": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "dart-dio",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "python": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "python",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "typescript-axios": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "typescript-axios",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "typescript-fetch": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "typescript-fetch",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "java": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "java",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "kotlin": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "kotlin",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "go": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "go",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "csharp": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "csharp",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "ruby": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "ruby",
            "-o", "{out_dir}"
        ],
        "ext": None
    },
    "swift5": {
        "cmd": [
            "openapi-generator-cli", "generate",
            "-i", "{schema}",
            "-g", "swift5",
            "-o", "{out_dir}"
        ],
        "ext": None
    }
}


def setup_django(settings_module: str):
    sys.path.insert(0, os.getcwd())
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", settings_module)
    
    import django
    django.setup()
    

def dump_openapi(api_module: str, api_var: str, out_file: Path):
    from django.test import RequestFactory
    from ninja.openapi.views import openapi_json
    
    mod = importlib.import_module(api_module)
    api = getattr(mod, api_var)
    
    req = RequestFactory().get("/api/openapi.json")
    resp = openapi_json(req, api)
    out_file.write_bytes(resp.content)
    

def generate_client(schema_path: Path, language: str, output: str):
    cfg = GENERATOR_CONFIG.get(language)

    cmd = [p.format(schema=str(schema_path), out=output, out_dir=output) for p in cfg["cmd"]]
    print(f"[LazyNinja] ▶ Running: {' '.join(cmd)}")
    proc = subprocess.run(cmd)
    if proc.returncode != 0:
        print(f"[LazyNinja] ❌ Generator failed.")
        sys.exit(proc.returncode)
    print(f"[LazyNinja] ✅ Client ({language}) generated at {output}")
    

def main():
    parser = argparse.ArgumentParser(
        prog="lazy-ninja",
        description=(
        "🌀 Lazy Ninja CLI\n\n"
        "Generate client code and SDKs from your Django + Ninja API schema.\n\n"
        "Supports generating frontend clients (e.g., TypeScript for React) as well as backend SDKs\n"
        "for server-to-server communication or internal services.\n\n"
        "Example usage:\n"
        "  lazy-ninja generate-client typescript \\\n"
        "    --settings myproject.settings \\\n"
        "    --api-module myproject.api \\\n"
        "    --output ./client.ts"
    ),
        epilog="🐛 Report issues at: https://github.com/AghastyGD/lazy-ninja/issues",
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    sub = parser.add_subparsers(dest="cmd")

    gen = sub.add_parser(
        "generate-client",
        help="Generate client code from OpenAPI schema",
        description=(
            "Generate client code from the OpenAPI schema exposed by your Django Ninja API.\n\n"
            "Supports multiple languages. You must provide your Django settings module and\n"
            "the path to the module where your `api = NinjaAPI()` instance is defined."
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )

    gen.add_argument(
        "language",
        choices=list(GENERATOR_CONFIG.keys()),
        help="Target language for client code (e.g. typescript, python)"
    )
    gen.add_argument(
        "--settings",
        required=True,
        help="Django settings module (e.g. myproject.settings)"
    )
    gen.add_argument(
        "--api-module",
        default="settings.api",
        help="Module path where your `api = NinjaAPI()` is defined (default: settings.api)"
    )
    gen.add_argument(
        "--output",
        default="./client",
        help="Output file or folder (e.g. ./client.ts)"
    )

    args = parser.parse_args()

    if not args.cmd:
        parser.print_help()
        sys.exit(1)

    setup_django(args.settings)
    schema_file = Path(".lazy_ninja_openapi.json")
    dump_openapi(args.api_module, "api", schema_file)
    generate_client(schema_file, args.language, args.output)
    schema_file.unlink(missing_ok=True)
    
    
if __name__ == "__main__":
    main()
    

    